/**
 * $Id: BagImpl.java,v 1.4 2006/07/07 11:38:53 uwe Exp $
 */
package de.fhw.data.bag;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.function.Predef;
import de.fhw.data.list.List;

/**
 * All default implementations for sets.
 * Most of these implementations use folds,
 * so the methods make a full traversal of the sets
 * This leeds to extremly inefficient operations.
 * 
 * @param <E> the Element type
 */

public abstract
class BagImpl<E>
implements Bag<E> {
	/**
	 * @return a single element set
	 */
	public
	Bag <E> single(E e) {
		return
		single(e,1);
	}
	
	/**
	 * build a single element bag
	 */
	public
	Bag <E> single(E e, int occ){
		return
		empty().insert(e,occ);
	}

	/**
	 * isEmpty() is implemented by card() == 0
	 */
	public
	boolean isEmpty() {
		return
		card() == 0;
	}

	/**
	 * isSubset implemented with all
	 */
	public
	boolean isSubbagOf(final Bag<E> s2){
		Fct<E,Boolean> isContained =
		new Fct<E,Boolean>() {
			public
			Boolean apply(E e) {
				return
				occurences(e) <= s2.occurences(e);
			}
		};
		return all(isContained);
	}

	/**
	 * isEqualTo implemented with card and contains
	 * 
	 */
	public
	boolean isEqualTo(Bag<E> s2) {
		if (this == s2)
			return true;
		
		if (card() != s2.card())
			return false;
		
		return 	contains(s2);
	}
	
	/**
	 * card is implemented with fold and counting
	 */
	public
	int card() {
		return
		fold(new Fct2<E,Integer,Integer>() {
			public
			Integer apply(E e, Integer r) {
				return
				r + 1;
			}
		}
		, 0
		);
	}
	
	/**
	 * contains is implemented with occurences
	 */
	public
	boolean contains(final E e) {
		return
		occurences(e) > 0;
	}

	/**
	 * contains implemented with isSubbagOf
	 */
	public
	boolean contains(Bag<E> s2) {
		return
		s2.isSubbagOf(this);
	}

	public E choose() {
		E e = choose1();
		if (e == null)
			throw
			new IllegalArgumentException("choose applied to empty bag");	
		return e;
	}

	protected
	abstract
	E choose1();
	
	/**
	 * insert an element
	 */
	public
	Bag <E> insert(E e) {
		return
		insert(e,1);
	}
	
	/**
	 * remove an element
	 */
	public
	Bag <E> remove(E e){
		return
		remove(e,1);
	}

	/**
	 * findMin implementation via fold
	 */
	public
	E findMin() {
		throw
		new UnsupportedOperationException("findMin not supported for bag implementation");
	}
	/**
	 * findMax implementation via fold
	 */
	public
	E findMax() {
		throw
		new UnsupportedOperationException("findMax not supported for bag implementation");
	}

	/**
	 * remove the minimum element
	 * 
	 * raise UnsupportedOperationException
	 */
	public
	Bag <E> removeMin() {
		throw
		new UnsupportedOperationException("removeMin not supported for Bag implementation");
	}
	
	/**
	 * remove the maximum element
	 * 
	 * raise UnsupportedOperationException
	 */
	public
	Bag <E> removeMax() {
		throw
		new UnsupportedOperationException("removeMax not supported for Bag implementation");
		
	}
	
	/**
	 * remove is implemented with fold, empty and insert
	 */
	public
	Bag <E> remove(final E e, int occ) {
		int occ1 = Math.max(occurences(e) - occ, 0);
		Bag <E> res =
			fold(new Fct2<E,Bag<E>,Bag<E>>() {
				public
				Bag <E> apply(E e1, Bag <E> r) {
					if (e.equals(e1))
						return
						r;
					return
					r.insert(e1);
				}
			}
			,empty()
			);
		if (occ1 == 0)
			return res;
		return res.insert(e,occ1);
	}
	
	/**
	 * union is implemented with fold and insert
	 */
	public
	Bag <E> union(Bag <E> b2) {
		return
		fold(new Fct2<E,Bag<E>,Bag<E>>() {
			public
			Bag<E> apply(E e, Bag<E> r) {
				return
				r.insert(e);
			}	
		}
		,b2);
	}
	
	/**
	 * difference is implemented with fold, and remove
	 */
	public
	Bag <E> difference(final Bag <E> b2) {
		return
		b2.fold(new Fct2<E,Bag<E>,Bag<E>>() {
			public
			Bag<E> apply(E e, Bag<E> r) {
				return
				r.remove(e);
			}
		}
		,this
		);
	}
	
	/**
	 * intersection is implemented with fold, isIn and insert
	 */
	public
	Bag <E> intersection(final Bag <E> b2) {
		return
		fold(new Fct2<E,Bag<E>,Bag<E>>() {
			public
			Bag<E> apply(E e, Bag<E> r) {
				if (r.occurences(e) < b2.occurences(e))
					return r.insert(e);
				return
				r;
			}
		}
		,empty()
		);
	}

	// TODO this declaration leads to error: method becomes ambiguous
	/**
	 * Ecipse bug: The method is ambiguous for the type ...
	 * 
	 * If bug is removed, rename this to foldMap and make it public,
	 * foldMap implementaions in subclasses, which call this method
	 * can then be removed
	 */
	protected
	<R>
	R eclipseBugFoldMap(final Fct2<R,R,R> combine, final Fct<E,R> map, R init) {
		Fct2<E,R,R> foldf =
			new Fct2<E,R,R>() {
			public
			R apply(E e, R r) {
				return
				combine.apply(map.apply(e),r);
			}
		};
		return
		fold(foldf,init);
	}

	// TODO this declaration leads to error: method becomes ambiguous
	/**
	 * Eclipse bug: The method is ambiguous for the type ...
	 * 
	 * If bug is removed, rename this to fold1Map and make it public,
	 * fold1Map implementaions in subclasses, which call this method
	 * can then be removed
	 */
	protected
	<R>
	R eclipseBugFold1Map(
			final Fct2<R,R,R> combine,
			final Fct<E,R> map,
			R emptyRes) {
		if (isEmpty())
			return emptyRes;
		
		E e = choose1();
		
		Fct2<E,R,R> foldf	=
			new Fct2<E,R,R>() {
			public
			R apply(E e, R r) {
				return
				combine.apply(map.apply(e),r);
			}
		};
		return
		remove(e).fold(foldf,map.apply(e));
	}

	/**
	 * general implementation of the simple map function
	 * with fold
	 */
	public
	Bag<E> map(final Fct<E,E> mf) {
		
		Fct2<E,Bag<E>,Bag<E>> f
		= new Fct2<E,Bag<E>,Bag<E>>() {
			public
			Bag<E> apply(E e, Bag<E> r) {
				return
				r.insert(mf.apply(e));
			}
		};
		return
		fold(f,empty());
	}

	/**
	 * general implementation of the more flexible map function
	 * with fold
	 */
	public
	<E2>
	Bag<E2> map(final Fct<E,E2> mf, Bag<E2> proto) {
		
		Fct2<E,Bag<E2>,Bag<E2>> f
		= new Fct2<E,Bag<E2>,Bag<E2>>() {
			public
			Bag<E2> apply(E e, Bag<E2> r) {
				return
				r.insert(mf.apply(e));
			}
		};
		return
		fold(f,proto.empty());
	}
	
	/**
	 * filter implemented with fold and insert
	 */
	public
	Bag <E> filter(final Fct<E,Boolean> filter) {
		Fct2<E,Bag<E>,Bag<E>> f
		= new Fct2<E,Bag<E>,Bag<E>>() {
			public
			Bag<E> apply(E e, Bag<E> res) {
				return
				filter.apply(e)
				? res.insert(e) : res;
			}
		};
		return
		fold(f,empty());
	}
	
	/**
	 * all quantor implemented with fold
	 */
	public
	Boolean all(Fct<E,Boolean> p) {
		return
		foldMap(Predef.Bool.and,p,true);
	}
	
	/**
	 * exists quantor implemented with fold
	 */
	public
	Boolean any(Fct<E,Boolean> p) {
		return
		foldMap(Predef.Bool.or,p,false);
	}

	/**
	 * toList conversion with fold
	 */
	public
	List<E> toList(List<E> proto) {
		Fct2<E,List<E>,List<E>> cv =
			new Fct2<E,List<E>,List<E>>() {
			public
			List<E> apply(E e, List<E> res) {
				return res.cons(e);
			}
		};
		return fold(cv,proto.empty());
	}
	
	/**
	 * equals implemented with isEqualTo
	 */
	@Override
	public
	boolean equals(Object o2) {
		if (o2 instanceof Bag)
			return this.isEqualTo((Bag<E>)o2);
		return false;
	}
	
	/**
	 * toString is implemented with isEmpty and fold1Map
	 */
	@Override
	public
	String toString() {
		if (isEmpty())
			return "{}";
		return
		"{"
		+ fold1Map(Predef.Str.join(", "), Predef.<E>show(), "")
		+ "}";
	}
}
