/**
 * $Id: Set.java,v 1.8 2006/07/07 11:38:54 uwe Exp $
 */
package de.fhw.data.set;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.list.List;
import de.fhw.data.util.Fingerprint;
import de.fhw.data.util.Invariant;
import de.fhw.data.util.Profile;

/**
 * The interface for all persistent set implementations
 * 
 */
public interface Set<E>
extends Invariant, Profile, Fingerprint {
	
	/**
	 * test on empty set
	 */
	boolean isEmpty();
	
	/**
	 * 
	 * @return number of elements in the set
	 */
	int card();
	
	/**
	 * search for an element
	 */
	boolean contains(E e);

	/**
	 * test whether this set contains a set s2
	 * <p>
	 * <code>s1.contains(s2)</code> is equivalent to
	 * <code>s2.isSubsetOf(s1)</code>
	 */
	boolean contains(Set<E> s2);

	/**
	 * test whether this set is a subset or equal to set s2
	 */
	boolean isSubsetOf(Set<E> s2);
	
	/**
	 * test whether this set is equal to another set
	 */
	boolean isEqualTo(Set<E> s2);

	/**
	 * build a new empty set
	 */
	Set <E> empty();

	/**
	 * build a single element set
	 */
	Set <E> single(E e);

	/**
	 * insert an element
	 */
	Set <E> insert(E e);
	
	/**
	 * remove an element
	 */
	Set <E> remove(E e);
	
	/**
	 * choose an arbitray element from a none empty set,
	 * if the set is empty, an IllegalArgumentException is thrown
	 */
	E choose();

	/**
	 * Only supported for ordered sets.
	 * If the set is empty, an IllegalArgumentException is thrown.
	 * If the set isn't ordered, an UnsupportedOperationException is thrown
	 * 
	 * @return the minimum element in a none empty set.
	 */
	E findMin();
	
	/**
	 * Only supported for ordered sets
	 * If the set is empty, an IllegalArgumentException is thrown.
	 * If the set isn't ordered, an UnsupportedOperationException is thrown
	 * 
	 * @return the maximum element in a none empty set
	 */
	E findMax();

	/**
	 * remove the minimum element,
	 * if the set is empty the operation returns the empty set
	 * 
	 * Only supported for ordered sets
	 */
	Set <E> removeMin();
	
	/**
	 * remove the maximum element,
	 * if the set is empty the operation returns the empty set
	 * 
	 * Only supported for ordered sets
	 */
	Set <E> removeMax();

	/**
	 * union of 2 sets
	 */	
	Set <E> union(final Set <E> s2);
	
	/**
	 * difference of 2 sets
	 */
	Set <E> difference(final Set <E> s2);
	
	/**
	 * intersection of 2 sets
	 */
	Set <E> intersection(final Set <E> s2);
	
	/**
	 * simple version of map: maps all elements,
	 * but works with a single set type
	 */
	Set <E> map(final Fct<E,E> mf);
	
	/**
	 * map all elements of a set and collect the results in a new set
	 */
	<E2>
	Set <E2> map(final Fct<E, E2> mf, Set<E2> proto);
	
	/**
	 * filter the elements of a set
	 * and remove all elements, for which the predicate doesn not hold
	 * @param filter the predicate
	 * @return the new possibly smaller set
	 */
	Set <E> filter(final Fct<E,Boolean> filter);
	
	/**
	 * fold all elements of a set into a new result
	 */
	<R>
	R fold(final Fct2<E,R,R> f, R init);
	
	/**
	 * map all elements of a set into a result type
	 * and combine these results, if the set is empty,
	 * the 3. parameter is taken as the result
	 * 
	 * @param <R> the result type 
	 * @param combine the combining function
	 * @param map the element mapping
	 * @param init the initial value for building the result,
	 *  it's theresult for the empty set
	 * @return the combined result
	 */
	<R>
	R foldMap(final Fct2<R,R,R> combine, final Fct<E,R> map, R init);
	
	/**
	 * map all elements of a set into a result type
	 * and combine these results, emptyRes is the result if the set is empty,
	 * otherwise this value is not used, especially for
	 * one element sets the result is the mapping of the single element, not the
	 * combination of the element and emptyRes.
	 * 
	 * this allows building of lists like "e1,e2,...,en" and "e1"
	 * where foldMap would build "e1,e2,...,en," and "e1,"
	 * 
	 * example application is the general toString method in de.fhw.data.set.SetImpl
	 * 
	 * @param <R> the result type 
	 * @param combine the combining function
	 * @param map the element mapping
	 * @param emptyRes the result for the empty set
	 * @return the combined result
	 */
	<R>
	R fold1Map(final Fct2<R,R,R> combine, final Fct<E,R> map, R emptyRes);
	
	/**
	 * test, whether for all elements in a set a predicate p holds,
	 * generalisation of logical and
	 */
	Boolean all(Fct<E,Boolean> p);
	
	/**
	 * test, whether for some elements in a set a predicate p holds,
	 * generalisation of logical or
	 */
	Boolean any(Fct<E,Boolean> p);
	
	/**
	 * conversion into a list
	 */
	List<E> toList(List<E> proto);
}
